"""Logger gRPC client."""
from datetime import datetime
from types import ModuleType
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Sequence, Union

import grpc
from fred_api.logger import service_logger_grpc_pb2_grpc, service_search_logger_history_grpc_pb2_grpc
from fred_api.logger.service_logger_grpc_pb2 import (CloseLogEntryRequest, CloseSessionRequest, CreateLogEntryRequest,
                                                     CreateSessionRequest, GetLogEntryTypesRequest, GetResultsRequest)
from fred_api.logger.service_search_logger_history_grpc_pb2 import GetLogEntryInfoRequest, ListLogEntriesRequest
from frgal import GrpcClient, GrpcDecoder
from google.protobuf.empty_pb2 import Empty as _google_protobuf_Empty

from .exceptions import LogEntryDoesNotExist, SessionDoesNotExist

# Structure for a log property
LogProperty = NamedTuple('LogProperty', [('value', str), ('children', Optional[Dict[str, str]])])

# Custom types
Properties = Dict[str, Union[LogProperty, Iterable[LogProperty]]]
References = Dict[str, Union[str, Iterable[str]]]
User = Union[str, int]


def _to_tuple(value: Any) -> Iterable:
    """Transform value into a tuple."""
    if isinstance(value, Iterable) and not isinstance(value, (str, LogProperty)):
        return tuple(value)
    return (value, )


class LoggerClient(GrpcClient):
    """Logger gRPC client."""

    grpc_service = 'Logger'

    def __init__(self, netloc: str, grpc_modules: Sequence[ModuleType] = None,
                 credentials: grpc.ChannelCredentials = None):
        """
        Initialize Logger GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            grpc_modules: Sequence of grpc modules. If None, default [service_logger_grpc_pb2_grpc] is used.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
        """
        decoder = GrpcDecoder()
        decoder.set_exception_decoder('log_entry_does_not_exist', LogEntryDoesNotExist)
        decoder.set_exception_decoder('session_does_not_exist', SessionDoesNotExist)
        super().__init__(netloc, decoder=decoder,
                         grpc_modules=grpc_modules if grpc_modules is not None else [service_logger_grpc_pb2_grpc],
                         credentials=credentials)

    def _update_log_request(self, request: Union[CreateLogEntryRequest, CloseLogEntryRequest], session_id: str = None,
                            content: str = None, properties: Properties = None,
                            references: References = None):
        """Update shared fields in create and close log entry requests."""
        if session_id:
            request.session_id.value = session_id
        if content:
            request.content.value = content
        properties = properties or {}
        for key, values in properties.items():
            values = _to_tuple(values)
            for value in values:
                if value.children:
                    request.properties[key].values.add(value=value.value, children=value.children)
                else:
                    request.properties[key].values.add(value=value.value)
        references = references or {}
        for key, ref_values in references.items():
            for value in _to_tuple(ref_values):
                request.references[key].references.add(value=value)

    def create_log_entry(self, service: str, log_entry_type: str, *, source_ip: str = None, session_id: str = None,
                         content: str = None, properties: Properties = None, references: References = None) -> str:
        """
        Send gRPC request to create log entry.

        Args:
            service: Name of the logger service.
            log_entry_type: Name of the log entry type.
            source_ip: IP address of the action source.
            session_id: Identifier of the log session.
            content: Input content.
            properties: Input properties.
            references: Input references.
        """
        request = CreateLogEntryRequest()
        request.service.value = service
        request.log_entry_type.value = log_entry_type
        if source_ip:
            request.source_ip.value = source_ip
        self._update_log_request(request, session_id, content, properties, references)

        return self.call(self.grpc_service, 'create_log_entry', request)

    def close_log_entry(self, log_entry_id: str, result: str, *, session_id: str = None, content: str = None,
                        properties: Properties = None, references: References = None):
        """
        Send gRPC request to close log entry.

        Args:
            log_entry_id: Identifier of the log entry to close.
            result: Name of the result.
            session_id: Identifier of the log session.
            content: Output content.
            properties: Output properties.
            references: Output references.
        """
        request = CloseLogEntryRequest()
        request.log_entry_id.value = log_entry_id
        request.result_name = result
        self._update_log_request(request, session_id, content, properties, references)

        return self.call(self.grpc_service, 'close_log_entry', request)

    def create_session(self, user_id: int, username: str) -> str:
        """
        Send gRPC request to create session.

        Args:
            user_id: User's identifier.
            username: User's username.
        """
        request = CreateSessionRequest()
        request.user_id.value = user_id
        request.username.value = username
        return self.call(self.grpc_service, 'create_session', request)

    def close_session(self, session_id: str):
        """
        Send gRPC request to create session.

        Args:
            session_id: Identifier of the session to close.
        """
        request = CloseSessionRequest()
        request.session_id.value = session_id
        return self.call(self.grpc_service, 'close_session', request)

    def get_log_entry_types(self, service: str):
        """
        Send gRPC request to load log entry types.

        Args:
            service: Identifier of the requested service.
        """
        request = GetLogEntryTypesRequest()
        request.service.value = service
        return self.call(self.grpc_service, 'get_log_entry_types', request)

    def get_services(self):
        """
        Send gRPC request to load all defined services.

        Args:
            No arguments
        """
        return self.call(self.grpc_service, 'get_services', _google_protobuf_Empty())

    def get_results(self, service: str):
        """
        Send gRPC request to load all defined results.

        Args:
            service: Identifier of the requested service.
        """
        request = GetResultsRequest()
        request.service.value = service
        return self.call(self.grpc_service, 'get_results', request)

    def get_object_reference_types(self):
        """
        Send gRPC request to get defined types of the object reference.

        Args:
            No arguments
        """
        return self.call(self.grpc_service, 'get_object_reference_types', _google_protobuf_Empty())


class SearchLoggerHistoryClient(GrpcClient):
    """Search Logger History gRPC client."""

    grpc_service = 'SearchLoggerHistory'

    def __init__(self, netloc: str, grpc_modules: Sequence[ModuleType] = None,
                 credentials: grpc.ChannelCredentials = None):
        """
        Initialize Search Logger History GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            grpc_modules: Sequence of grpc modules. If None,
             default [service_search_logger_history_grpc_pb2_grpc] is used.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
        """
        super().__init__(
            netloc,
            decoder=GrpcDecoder(),
            grpc_modules=grpc_modules if grpc_modules is not None else [service_search_logger_history_grpc_pb2_grpc],
            credentials=credentials
        )

    def list_log_entries(self, service: str, timestamp_from: datetime, timestamp_to: datetime, limit: int, *,
                         log_entry_types: List[str] = None, user: User = None,
                         reference_type: str = None, reference_value: str = None) -> List[str]:
        """Return a list of filtered log entry identifiers."""
        """
        Send gRPC request to get log entry info.

        Args:
            service: Name of the logger service.
            timestamp_from: Begin of the time range filter applied to search.
            timestamp_to: End of the time range filter applied to search.
            limit: Number of the maximum count of results.
            log_entry_types: List of log entry type names.
            user: User name or user_id
            reference_type: Type of object referenced in searched log entries.
            reference_value: Object identifier referenced in searched log entries.

        Returns:
            List: list of found log entry identifiers.
        """
        request = ListLogEntriesRequest()
        # required parameters
        request.service.value = service
        request.timestamp_from.FromDatetime(timestamp_from)
        request.timestamp_to.FromDatetime(timestamp_to)
        request.count_limit = limit

        for log_entry_type in (log_entry_types or []):
            request.log_entry_types.add(value=log_entry_type)
        if user:
            if isinstance(user, int):
                request.user_id.value = user
            else:
                request.username.value = user
        if reference_type:
            request.reference.type.value = reference_type
        if reference_value:
            request.reference.value.value = reference_value

        # The call may return `None`.
        return self.call(self.grpc_service, 'list_log_entries', request) or []

    def get_log_entry_info(self, log_entry_id: str) -> Dict[str, Any]:
        """
        Send gRPC request to get log entry info.

        Args:
            log_entry_id: Identifier of the log entry to retrieve details about.

        Returns:
            dictionary: details of the requested log entry.
        """
        request = GetLogEntryInfoRequest()
        request.log_entry_id.value = log_entry_id
        return self.call(self.grpc_service, 'get_log_entry_info', request)
