"""Tests for grpc module."""
from datetime import datetime, timezone
from typing import Any, Dict
from unittest import TestCase
from unittest.mock import call, sentinel

from fred_api.logger.log_entry_common_types_pb2 import LogEntryDoesNotExist as GrpcLogEntryDoesNotExist
from fred_api.logger.service_logger_grpc_pb2 import (CloseLogEntryReply, CloseLogEntryRequest, CloseSessionReply,
                                                     CloseSessionRequest, CreateLogEntryReply, CreateLogEntryRequest,
                                                     CreateSessionReply, CreateSessionRequest, GetLogEntryTypesReply,
                                                     GetLogEntryTypesRequest, GetObjectReferenceTypesReply,
                                                     GetResultsReply, GetResultsRequest, GetServicesReply)
from fred_api.logger.service_search_logger_history_grpc_pb2 import (GetLogEntryInfoReply, GetLogEntryInfoRequest,
                                                                    ListLogEntriesReply, ListLogEntriesRequest)
from frgal.utils import TestClientMixin
from google.protobuf.empty_pb2 import Empty as _google_protobuf_Empty

from grill.exceptions import LogEntryDoesNotExist, SessionDoesNotExist
from grill.grpc import LoggerClient, LogProperty, SearchLoggerHistoryClient


class TestLoggerClient(TestClientMixin, LoggerClient):
    """Testing version of a LoggerClient."""


class LoggerClientTest(TestCase):
    def setUp(self):
        self.client = TestLoggerClient(sentinel.netloc)

    def test_grpc_modules(self):
        client = TestLoggerClient(sentinel.netloc, grpc_modules=[])
        self.assertEqual(client.grpc_modules, [])

    def _test_create_log_entry(self, request: CreateLogEntryRequest, **kwargs):
        reply = CreateLogEntryReply()
        reply.data.log_entry_id.value = 'FourtyTwo'
        self.client.mock.return_value = reply

        entry_id = self.client.create_log_entry('Universe', 'Answer', **kwargs)

        self.assertEqual(entry_id, 'FourtyTwo')
        # Finish and check request
        request.service.value = 'Universe'
        request.log_entry_type.value = 'Answer'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/create_log_entry', timeout=None)])

    def test_create_log_entry_minimal(self):
        # Test create_log_entry with minimum arguments
        request = CreateLogEntryRequest()
        self._test_create_log_entry(request)

    def test_create_log_entry_source_ip(self):
        request = CreateLogEntryRequest()
        request.source_ip.value = '256.0.0.1'
        self._test_create_log_entry(request, source_ip='256.0.0.1')

    def test_create_log_entry_session_id(self):
        request = CreateLogEntryRequest()
        request.session_id.value = 'Life'
        self._test_create_log_entry(request, session_id='Life')

    def test_create_log_entry_content(self):
        request = CreateLogEntryRequest()
        request.content.value = 'TheQuestion'
        self._test_create_log_entry(request, content='TheQuestion')

    def test_create_log_entry_properties_simple(self):
        request = CreateLogEntryRequest()
        request.properties['quality'].values.add(value='absolute')
        self._test_create_log_entry(request, properties={'quality': LogProperty('absolute', None)})

    def test_create_log_entry_properties_children(self):
        request = CreateLogEntryRequest()
        request.properties['quality'].values.add(value='absolute', children={'control': 'yes'})
        self._test_create_log_entry(request, properties={'quality': LogProperty('absolute', {'control': 'yes'})})

    def test_create_log_entry_properties_list(self):
        request = CreateLogEntryRequest()
        request.properties['quality'].values.add(value='absolute')
        request.properties['quality'].values.add(value='relative')
        self._test_create_log_entry(
            request, properties={'quality': [LogProperty('absolute', None), LogProperty('relative', None)]})

    def test_create_log_entry_references_simple(self):
        request = CreateLogEntryRequest()
        request.references['footnote'].references.add(value='196')
        self._test_create_log_entry(request, references={'footnote': '196'})

    def test_create_log_entry_references_list(self):
        request = CreateLogEntryRequest()
        request.references['footnote'].references.add(value='196')
        request.references['footnote'].references.add(value='197')
        self._test_create_log_entry(request, references={'footnote': ['196', '197']})

    def _test_close_log_entry(self, request: CloseLogEntryRequest, **kwargs):
        self.client.mock.return_value = CloseLogEntryReply()

        self.client.close_log_entry('FourtyTwo', 'Success', **kwargs)

        # Finish and check request
        request.log_entry_id.value = 'FourtyTwo'
        request.result_name = 'Success'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/close_log_entry', timeout=None)])

    def test_close_log_entry_minimal(self):
        # Test close_log_entry with minimum arguments
        request = CloseLogEntryRequest()
        self._test_close_log_entry(request)

    def test_close_log_entry_session_id(self):
        request = CloseLogEntryRequest()
        request.session_id.value = 'Life'
        self._test_close_log_entry(request, session_id='Life')

    def test_close_log_entry_content(self):
        request = CloseLogEntryRequest()
        request.content.value = 'TheQuestion'
        self._test_close_log_entry(request, content='TheQuestion')

    def test_close_log_entry_properties_simple(self):
        request = CloseLogEntryRequest()
        request.properties['quality'].values.add(value='absolute')
        self._test_close_log_entry(request, properties={'quality': LogProperty('absolute', None)})

    def test_close_log_entry_properties_children(self):
        request = CloseLogEntryRequest()
        request.properties['quality'].values.add(value='absolute', children={'control': 'yes'})
        self._test_close_log_entry(request, properties={'quality': LogProperty('absolute', {'control': 'yes'})})

    def test_close_log_entry_properties_list(self):
        request = CloseLogEntryRequest()
        request.properties['quality'].values.add(value='absolute')
        request.properties['quality'].values.add(value='relative')
        self._test_close_log_entry(
            request, properties={'quality': [LogProperty('absolute', None), LogProperty('relative', None)]})

    def test_close_log_entry_references_simple(self):
        request = CloseLogEntryRequest()
        request.references['footnote'].references.add(value='196')
        self._test_close_log_entry(request, references={'footnote': '196'})

    def test_close_log_entry_references_list(self):
        request = CloseLogEntryRequest()
        request.references['footnote'].references.add(value='196')
        request.references['footnote'].references.add(value='197')
        self._test_close_log_entry(request, references={'footnote': ['196', '197']})

    def test_close_log_entry_error(self):
        reply = CloseLogEntryReply()
        reply.exception.log_entry_does_not_exist.CopyFrom(GrpcLogEntryDoesNotExist())
        self.client.mock.return_value = reply

        with self.assertRaises(LogEntryDoesNotExist):
            self.client.close_log_entry('FourtyTwo', 'Success')

        # Check call
        request = CloseLogEntryRequest()
        request.log_entry_id.value = 'FourtyTwo'
        request.result_name = 'Success'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/close_log_entry', timeout=None)])

    def test_create_session(self):
        reply = CreateSessionReply()
        reply.data.session_id.value = 'Life'
        self.client.mock.return_value = reply

        self.client.create_session(4242, 'Arthur Dent')

        # Finish and check request
        request = CreateSessionRequest()
        request.user_id.value = 4242
        request.username.value = 'Arthur Dent'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/create_session', timeout=None)])

    def test_close_session(self):
        self.client.mock.return_value = CloseSessionReply()

        self.client.close_session('Life')

        # Finish and check request
        request = CloseSessionRequest()
        request.session_id.value = 'Life'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/close_session', timeout=None)])

    def test_close_session_error(self):
        reply = CloseSessionReply()
        reply.exception.session_does_not_exist.CopyFrom(CloseSessionReply.Exception.SessionDoesNotExist())
        self.client.mock.return_value = reply

        with self.assertRaises(SessionDoesNotExist):
            self.client.close_session('Life')

        # Check call
        request = CloseSessionRequest()
        request.session_id.value = 'Life'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/close_session', timeout=None)])

    def test_get_log_entry_types(self):
        reply = GetLogEntryTypesReply()
        reply.data.types.add(value='hello')
        reply.data.types.add(value='bye')
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_log_entry_types(service='Earth'), ['hello', 'bye'])

        # Check call
        request = GetLogEntryTypesRequest()
        request.service.value = 'Earth'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/get_log_entry_types', timeout=None)])

    def test_get_services(self):
        reply = GetServicesReply()
        reply.data.services.add(value='Alpha')
        reply.data.services.add(value='Beta')
        reply.data.services.add(value='Gama')
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_services(), ['Alpha', 'Beta', 'Gama'])

        # Check call
        request = _google_protobuf_Empty()
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/get_services', timeout=None)])

    def test_get_results(self):
        reply = GetResultsReply()
        reply.data.result_code_map['ok'] = 0
        reply.data.result_code_map['error'] = 1
        reply.data.result_code_map['mega-error'] = 9999
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_results(service='Earth'), {'ok': 0, 'error': 1, 'mega-error': 9999})

        # Check call
        request = GetResultsRequest()
        request.service.value = 'Earth'
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/get_results', timeout=None)])

    def test_get_object_reference_types(self):
        reply = GetObjectReferenceTypesReply()
        reply.data.object_types.add(value='number')
        reply.data.object_types.add(value='boolean')
        reply.data.object_types.add(value='string')
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_object_reference_types(), ['number', 'boolean', 'string'])

        # Check call
        request = _google_protobuf_Empty()
        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.Logger/get_object_reference_types', timeout=None)])


class TestSearchLoggerHistoryClient(TestClientMixin, SearchLoggerHistoryClient):
    """Testing version of a SearchLoggerHistoryClient."""


class SearchLoggerHistoryClientTest(TestCase):
    """Test SearchLoggerHistoryClient class."""

    def setUp(self):
        self.client = TestSearchLoggerHistoryClient(sentinel.netloc)

    def test_grpc_modules(self):
        client = TestSearchLoggerHistoryClient(sentinel.netloc, grpc_modules=[])
        self.assertEqual(client.grpc_modules, [])

    def test_list_log_entries_empty(self):
        reply = ListLogEntriesReply()
        self.client.mock.return_value = reply
        self.assertEqual(self.client.list_log_entries(service='Globe',
                                                      timestamp_from=datetime(1988, 11, 25, 18, 0),
                                                      timestamp_to=datetime(1988, 12, 24, 18, 30),
                                                      limit=100,
                                                      log_entry_types=['RimmersDiary'],
                                                      user="Albert",
                                                      reference_type='dog', reference_value='42'), [])

        # Check call
        request = ListLogEntriesRequest()
        request.service.value = 'Globe'
        request.timestamp_from.FromDatetime(datetime(1988, 11, 25, 18, 0))
        request.timestamp_to.FromDatetime(datetime(1988, 12, 24, 18, 30))
        request.count_limit = 100
        request.log_entry_types.add(value='RimmersDiary')
        request.username.value = "Albert"
        request.reference.type.value = 'dog'
        request.reference.value.value = '42'

        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.SearchLoggerHistory/list_log_entries',
                               timeout=None)])

    def _test_list_log_entries(self, filters: Dict[str, Any], request: ListLogEntriesRequest):
        reply = ListLogEntriesReply()
        reply.data.log_entry_ids.add(value='GazpachoSoupDay')
        self.client.mock.return_value = reply

        self.assertEqual(self.client.list_log_entries(service='Globe',
                                                      timestamp_from=datetime(1988, 11, 25, 18, 0),
                                                      timestamp_to=datetime(1988, 12, 24, 18, 30),
                                                      limit=100, **filters), ['GazpachoSoupDay'])

        # Check call
        request.service.value = 'Globe'
        request.timestamp_from.FromDatetime(datetime(1988, 11, 25, 18, 0))
        request.timestamp_to.FromDatetime(datetime(1988, 12, 24, 18, 30))
        request.count_limit = 100

        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.SearchLoggerHistory/list_log_entries',
                               timeout=None)])

    def test_list_log_entries(self):
        request = ListLogEntriesRequest()
        self._test_list_log_entries({}, request)

    def test_list_log_entries_username(self):
        request = ListLogEntriesRequest()
        request.username.value = "Albert"
        self._test_list_log_entries({'user': "Albert"}, request)

    def test_list_log_entries_user_id(self):
        request = ListLogEntriesRequest()
        request.user_id.value = 1001
        self._test_list_log_entries({'user': 1001}, request)

    def test_list_log_entries_log_entry_types(self):
        request = ListLogEntriesRequest()
        request.log_entry_types.add(value='RimmersDiary')
        self._test_list_log_entries({'log_entry_types': ['RimmersDiary']}, request)

    def test_list_log_entries_reference(self):
        request = ListLogEntriesRequest()
        request.reference.type.value = 'dog'
        request.reference.value.value = '42'
        self._test_list_log_entries({'reference_type': 'dog', 'reference_value': '42'}, request)

    def test_get_log_entry_info(self):
        reply = GetLogEntryInfoReply()
        reply.data.log_entry_info.time_begin.FromDatetime(datetime(1988, 11, 25, 18))
        self.client.mock.return_value = reply

        result = self.client.get_log_entry_info('GazpachoSoupDay')

        self.assertEqual(result['time_begin'], datetime(1988, 11, 25, 18, tzinfo=timezone.utc))

        # Check call
        request = GetLogEntryInfoRequest()
        request.log_entry_id.value = 'GazpachoSoupDay'

        self.assertEqual(self.client.mock.mock_calls,
                         [call(request, method='/Fred.Logger.Api.SearchLoggerHistory/get_log_entry_info',
                               timeout=None)])
